from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import Paginator, EmptyPage
from django.db.models import Q

from worf.casing import camel_to_snake
from worf.exceptions import HTTP420
from worf.views.base import AbstractBaseAPI
from worf.views.create import CreateAPI


class ListAPI(AbstractBaseAPI):
    results_per_page = 25
    lookup_url_kwarg = "id"  # default incase lookup_field is set
    filters = {}
    ordering = []
    filter_fields = None
    search_fields = None
    sort_fields = None
    queryset = None
    q_objects = Q()
    count = 0
    page_num = 1
    num_pages = 1

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(self.filters, dict):
            raise ImproperlyConfigured(f"{self.codepath}.filters must be type: dict")
        if not isinstance(self.ordering, list):
            raise ImproperlyConfigured(f"{self.codepath}.ordering must be type: list")
        if self.filter_fields is not None and not isinstance(self.filter_fields, list):
            raise ImproperlyConfigured(
                f"{self.codepath}.filter_fields must be type: list"
            )
        if self.search_fields is not None and not isinstance(self.search_fields, dict):
            raise ImproperlyConfigured(
                f"{self.codepath}.search_fields must be type: dict"
            )
        if self.sort_fields is not None and not isinstance(self.sort_fields, list):
            raise ImproperlyConfigured(
                f"{self.codepath}.sort_fields must be type: list"
            )

    def _set_base_lookup_kwargs(self):
        # Filters set directly on the class
        self.lookup_kwargs.update(self.filters)

        # Filters set via URL
        if hasattr(self, "lookup_field") and hasattr(self, "lookup_url_kwarg"):
            self.lookup_kwargs.update(
                {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}
            )

        self.validate_lookup_field_values()

    def set_search_lookup_kwargs(self):
        """
        Set generic lookup kwargs based on q and additional GET params.

        For more advanced search use cases, override this method and pass GET
        with any remaining params you want to use classic django filters for.
        """
        if self.search_fields is None:
            """If self.search_fields is not set, we don't allow search."""
            return

        self.set_bundle_from_querystring()
        # Whatever is not q or page as a querystring param will
        # be used for key-value search.
        search_string = self.bundle.get("q", False)
        self.bundle.pop("page", None)
        self.bundle.pop("p", None)
        self.bundle.pop("q", None)

        if search_string and len(search_string):
            the_ors = self.search_fields.get("or", [])
            for attr in the_ors:
                kwarg = {attr + "__icontains": search_string.strip()}
                self.q_objects.add(Q(**kwarg), Q.OR)

            the_ands = self.search_fields.get("ands", [])
            for attr in the_ands:
                kwarg = {attr + "__icontains": search_string.strip()}

        if not self.bundle:
            return

        for key in self.bundle.keys():
            if self.filter_fields is not None and key not in self.filter_fields:
                continue

            if self.get_field_type(key) == "ManyToManyField":
                if not isinstance(self.bundle[key], list):
                    # TODO simplify this when we move to POST for search.
                    # We do type coersion in set_bundle_from_querystring,
                    # so we need to restore the type of M2M fields as list.
                    self.bundle[key] = [self.bundle[key]]

                self.validate_bundle(key)

            # If we get a list, only support integer values, and skip bundle validation.
            if isinstance(self.bundle[key], list):
                if not all(isinstance(x, int) for x in self.bundle[key]):
                    self.coerce_array_of_integers(key)  # raises 422 if failure
                self.lookup_kwargs.update({f"{key}__in": self.bundle[key]})
                continue

            self.validate_bundle(key)
            self.lookup_kwargs.update({key: self.bundle[key]})

    def get_queryset(self):
        return (self.queryset or self.model.objects).all()

    def get_processed_queryset(self):
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~  URL Kwargs
        # If these aren't reset they are polluted by cache (somehow)
        self.lookup_kwargs = {}
        self.q_objects = Q()
        result_set = None

        self._set_base_lookup_kwargs()
        self.set_search_lookup_kwargs()

        order_by = self.ordering
        if self.request.GET.get("sort"):
            sort = self.parse_sort(self.request.GET.get("sort"))
            if sort.lstrip("-") in self.sort_fields:
                order_by = [sort]

        try:
            result_set = (
                self.get_queryset()
                .filter(self.q_objects, **self.lookup_kwargs)
                .order_by(*order_by)
                .distinct()
            )
        except TypeError as e:
            if settings.DEBUG:
                raise HTTP420(f"Error, {self.lookup_kwargs}, {e.__cause__}")
            raise e

        return result_set

    def paginated_results(self):
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ PAGINATION
        queryset = self.get_processed_queryset()
        request = self.request

        if settings.DEBUG:
            self.query = str(queryset.query)

        if self.results_per_page is None:
            return queryset

        paginator = Paginator(queryset, self.results_per_page)

        self.page_num = int(request.GET.get("page") or request.GET.get("p") or 1)
        if self.page_num < 1:
            self.page_num = 1

        self.num_pages = paginator.num_pages
        self.count = paginator.count

        try:
            return paginator.page(self.page_num)
        except EmptyPage:
            return []

    def parse_sort(self, sort):
        prefix = "-" if sort[0] == "-" else ""
        return prefix + "__".join(map(camel_to_snake, sort.lstrip("-").split(".")))

    def serialize(self):
        serializer = self.get_serializer()

        payload = {
            str(self.name): [
                serializer.read(instance) for instance in self.paginated_results()
            ]
        }

        if self.results_per_page:
            payload.update(
                {
                    "pagination": dict(
                        count=self.count,
                        pages=self.num_pages,
                        page=self.page_num,
                    )
                }
            )

        if not settings.DEBUG:
            return payload

        if not hasattr(self, "lookup_kwargs"):
            # Debug throws an error in the event there are no lookup_kwargs
            self.lookup_kwargs = {}

        payload.update(
            {
                "debug": {
                    "api_method": self.api_method,
                    "bundle": self.bundle,
                    "lookup_kwargs": self.lookup_kwargs,
                    "query": self.query,
                    "q_objs": str(self.q_objects),
                    "serializer": str(serializer),
                }
            }
        )

        return payload


class ListCreateAPI(ListAPI, CreateAPI):
    pass



import json
import warnings

from random import choice
from string import ascii_uppercase

from django.utils import timezone

from worf.casing import camel_to_snake, snake_to_camel


def generate_endpoint_tests(TestCase, API, instance, uri, patch=False, overrides={}):
    """
    Generate tests against an API, instance and uri.

    @param API the API Class to be tested
    @param instance: the model instance (usually created in test setup)
    @param uri: the path to the endpoint
    @param patch: Optionally test patch
    @param overrides: Optionally override values in the bundle factory
    """

    # TODO - This only works wih the model method serialization.
    # if API.serializer is not None:

    bundle = {
        **bundle_factory(instance, f"{API.api_method}_update_fields", overrides),
        **overrides,
    }
    response = TestCase.client.get(uri)
    TestCase.assertEqual(response.status_code, 200)

    if patch is False:
        return

    response = TestCase.client.patch(uri, bundle, content_type="application/json")
    TestCase.assertEqual(
        "20",  # assert 200-ish
        str(response.status_code)[0:2],
        response.content.decode("UTF-8"),
    )
    assertion_factory(TestCase, instance, bundle, overrides)


def assertion_factory(TestCase, instance, bundle, overrides={}):
    """
    Automatically test that a django model matches bundle content.

    @param TestCase: A TestCase class instance, typically `self`
    @param instance: A model instance
    @param bundle: A dictionary, potentially created by bundle_factory
    """
    # TODO consider extending the django test client with these methods. 👇
    instance.refresh_from_db()
    for key in bundle.keys():
        field = camel_to_snake(key)

        if field in overrides:
            continue

        if instance._meta.get_field(field).get_internal_type() in ["ManyToManyField"]:
            result = [x.pk for x in getattr(instance, field).all()]
            TestCase.assertCountEqual(
                result,
                bundle[key],
                msg=f"M2M {key} not set. Bundle was: \n\n {bundle}. Instance saved: \n\n {result}",
            )
            continue

        result = getattr(instance, field)
        TestCase.assertEqual(
            result,
            bundle[key],
            msg=f"{key} not set. Bundle was: \n\n {bundle}. Instance saved: \n\n {result}",
        )


def bundle_factory(instance, serializer="api_update_fields", overrides={}):
    # TODO consider extending the django test client with these methods.
    # TODO handle Choices fields; otherwise we may pass invalid values
    """
    Return a valid python dictionary for api testing.

    @param instance the model instance we are testing against
    @param serializer (optional) the model method that sets update_fields

    Limitiations:
        - Does not generate nor include any FK's, 1:1's, M2M fields
        - Only sets up valid data. Should always 200
    """
    bundle = dict()

    sequence = 0
    for field in getattr(instance, serializer)():
        sequence += 1
        type = instance._meta.get_field(field).get_internal_type()

        if field in overrides:
            continue

        if type in ["OneToOneField", "ForeignKey", "ManyToManyField"]:
            # TODO handle m2ms
            continue

        value = None
        if type in ["CharField", "TextField", "EmailField", "SlugField"]:
            if type == "TextField":
                value = "something {}".format(sequence)
            elif type == "EmailField":
                value = "test{}@example.com".format(sequence)
            else:
                max_len = instance._meta.get_field(field).max_length
                value = "".join(choice(ascii_uppercase) for i in range(max_len))

        elif type in ["IntegerField", "PositiveIntergerField", "SmallIntegerField"]:
            value = sequence

        elif type in ["BooleanField"]:
            value = True if sequence % 2 == 0 else False

        elif type in ["DateField"]:
            value = timezone.now().date()

        # TODO support JSONField

        if value is None:
            raise NotImplementedError(f"{type} is not implemented")
            # TODO
            # FileField
            # UUIDField

        bundle[snake_to_camel(field)] = value

    return bundle


def deserialize(response):
    # TODO consider extending the django test client with these methods. 👆
    return json.loads(response.content.decode("UTF-8"))


def verify_model_interface(instance, api="api"):
    """
    Ensure that any editable snake case field has a corresponding camel.

    Right now it's just dumb luck that they match, there's no way to test this
    until now.

    TODO adapt this to work with Serializer
    """
    if not hasattr(instance, f"{api}_update_fields"):
        warnings.warn(
            f"{instance.__class__.__name__} does not have a method for {api}_update_fields"
        )
        return True

    editable_fields = getattr(instance, f"{api}_update_fields")()

    interface = getattr(instance, api)().keys()

    for snake in editable_fields:
        assert hasattr(
            instance, snake
        ), f"{instance.__class__.__name__} does not have field {snake}"
        camel = snake_to_camel(snake)
        assert (
            camel in interface
        ), f"{camel} not present in {instance.__class__.__name__}.{api}"



from django.core.exceptions import ValidationError
from django.contrib.auth.models import User

from worf.permissions import PublicEndpoint
from worf.serializers import Serializer
from worf.views import DetailUpdateAPI

from tests.models import DummyModel


class DummyAPI(DetailUpdateAPI):
    model = DummyModel
    permissions = [PublicEndpoint]

    def validate_phone(self, value):
        try:
            assert value == "(555) 555-5555"
        except AssertionError:
            raise ValidationError("{value} is not a valid phone number")
        return "+5555555555"


class UserSerializer(Serializer):
    def read(self, model):
        return dict(
            username=model.username,
            lastLogin=model.last_login,
            dateJoined=model.date_joined,
            email=model.email,
        )

    def write(self):
        return [
            "username",
            "email",
        ]


class UserAPI(DetailUpdateAPI):
    model = User
    serializer = UserSerializer
    permissions = [PublicEndpoint]



from django.core.exceptions import ImproperlyConfigured

from worf.views.base import AbstractBaseAPI
from worf.shortcuts import get_instance_or_http404


class DetailAPI(AbstractBaseAPI):
    lookup_field = "id"
    lookup_url_kwarg = "id"
    instance = None

    def get(self, request, *args, **kwargs):
        return self.render_to_response()

    def serialize(self):
        """Return the model api, used for responses."""
        serializer = self.get_serializer()
        payload = serializer.read(self.get_instance())
        if not isinstance(payload, dict):
            raise ImproperlyConfigured(f"{serializer} did not return a dictionary")
        return payload

    def get_instance(self):
        # TODO support multiple lookup_fields
        self.lookup_kwargs = {self.lookup_field: self.kwargs[self.lookup_url_kwarg]}

        self.validate_lookup_field_values()

        if self.instance is None:
            self.instance = get_instance_or_http404(self.model, **self.lookup_kwargs)

        return self.instance

    def validate_and_update(self):
        """
        Update all fields passed in from json.

        Step 1: Validate
        Step 2: Update
        """
        instance = self.get_instance()
        fields = self.bundle.keys()

        for field in fields:
            self.validate_bundle(field)
            setattr(instance, field, self.bundle[field])

        instance.save()
        instance.refresh_from_db()


class DetailUpdateAPI(DetailAPI):
    def patch(self, request, *args, **kwargs):
        self.validate_and_update()
        return self.get(request)



import json
import types
import warnings

from urllib.parse import parse_qs

from django.conf import settings
from django.core.exceptions import (
    ImproperlyConfigured,
    ObjectDoesNotExist,
    ValidationError,
)
from django.http import JsonResponse
from django.middleware.gzip import GZipMiddleware
from django.views import View
from django.views.decorators.cache import never_cache
from django.utils.decorators import method_decorator

from worf.casing import camel_to_snake, whitespace_to_camel
from worf.exceptions import HTTP_EXCEPTIONS, HTTP404, HTTP422, PermissionsException
from worf.serializers import LegacySerializer
from worf.validators import ValidationMixin

gzip_middleware = GZipMiddleware()


@method_decorator(never_cache, name="dispatch")
class APIResponse(View):
    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"
        return super().__init__(*args, **kwargs)

    def serialize(self):
        raise NotImplementedError

    def render_to_response(self, data=None, status_code=None):
        payload = data if data is not None else self.serialize()

        if payload is None:
            msg = f"{self.codepath} did not pass an object to "
            msg += "render_to_response, nor did its serializer method"
            raise ImproperlyConfigured(msg)

        response = JsonResponse(payload)
        # except TypeError:
        # TODO add something meaningful to the stack trace

        if status_code is not None:  # needs tests
            response.status_code = status_code

        return gzip_middleware.process_response(self.request, response)


class AbstractBaseAPI(APIResponse, ValidationMixin):
    model = None
    permissions = []
    api_method = "api"
    serializer = None
    payload_key = None

    def __init__(self, *args, **kwargs):
        self.codepath = f"{self.__module__}.{self.__class__.__name__}"

        if self.model is None:
            raise ImproperlyConfigured(f"Model is not set on {self.codepath}")

        if not isinstance(self.permissions, list):
            raise ImproperlyConfigured(
                f"{self.codepath}.permissions must be type: list"
            )

        for method in ["post", "patch", "put", "delete"]:
            if method in dir(self) and not len(self.permissions):
                warnings.warn(
                    "\n{} method allowed on {} without permissions.".format(
                        method.upper(),
                        self.codepath,
                    ),
                )

        for method in self.permissions:
            # append authorization functions to this class
            setattr(self, method.__name__, types.MethodType(method, self))

        super().__init__(*args, **kwargs)

    @property
    def name(self):
        if isinstance(self.payload_key, str):
            return self.payload_key
        return whitespace_to_camel(self.model._meta.verbose_name_plural)

    def _check_permissions(self):
        """Return a permissions exception when in debug mode instead of 404."""
        for method in self.permissions:
            permission_func = getattr(self, method.__name__)
            response = permission_func(self.request)
            if response == 200:
                continue

            if settings.DEBUG:
                raise PermissionsException(
                    "Permissions function {}.{} returned {}. You'd normally see a 404 here but DEBUG=True.".format(
                        method.__module__, method.__name__, response
                    )
                )

            try:
                raise response
            except TypeError:
                raise ImproperlyConfigured(
                    "Permissions function {}.{} must return 200 or an HTTPException".format(
                        method.__module__, method.__name__
                    )
                )

    def _assemble_bundle_from_request_body(self):

        if self.request.content_type == "multipart/form-data":
            # Avoid RawPostDataException
            # TODO investigate why test did not catch this error
            raw_bundle = {}

        elif self.request.body:  # and self.request.body != b'--BoUnDaRyStRiNg--\r\n':
            try:
                raw_bundle = json.loads(self.request.body)
            except json.decoder.JSONDecodeError:
                # print("\n\n~~~~~~~~~~~~~~~~~~~~~~", self.request.body, '\n\n')
                raw_bundle = {}
        else:
            raw_bundle = {}

        self.set_bundle(raw_bundle)

    def _get_lookup_field_type(self, field):

        related = field.find("__")
        """Support one level of related field reference."""
        if related != -1:
            model = field[:related]
            target = field[related + 2 :]

            if target in ["gt", "lt", "contains", "startswith", "gte", "lte"]:
                return False

            if target.find("__") != -1:
                return False

            return (
                self.model._meta.get_field(model)
                .related_model._meta.get_field(target)
                .get_internal_type()
            )
            # TODO if there is another reference, recurse

        return self.model._meta.get_field(field).get_internal_type()

    def get_serializer(self):
        if self.serializer:
            return self.serializer()
        if self.api_method:
            return LegacySerializer(self.model, self.api_method)
        msg = f"{self.__name__}.get_serializer() did not return a serializer"
        raise ImproperlyConfigured(msg)

    def validate_lookup_field_values(self):
        # todo check for each lookup kwarg
        for field, url_kwarg in self.lookup_kwargs.items():

            lookup_field_type = self._get_lookup_field_type(field)

            if lookup_field_type == "UUIDField":
                self.validate_uuid(url_kwarg)
            elif lookup_field_type in [
                "IntegerField",
                "SmallIntegerField",
                "PositiveIntergerField",
                "ForeignKey",
            ]:
                self.validate_int(url_kwarg)

    def set_bundle_from_querystring(self):
        # parse_qs gives us a dictionary where all values are lists
        qs = parse_qs(self.request.META["QUERY_STRING"])

        # TODO: TLDR; Switch to POST for search instead of GET/querystring params.
        # we want to preserve strings wherever there are not duplicate keys
        # Step through the list and construct a dictionary for all fields
        # that are not duplicated

        # fundamentally all urlparams are treated as arrays natively.
        # we can't enforce type coersion here...

        # we can't assume everything is an array or everything is not an array
        # when it's a querystring

        raw_bundle = {}
        for key, value in qs.items():
            raw_bundle[key] = value[0] if len(value) == 1 else value

        self.set_bundle(raw_bundle)

    def set_bundle(self, raw):
        self.bundle = {}
        if not raw:
            return  # No need to loop or set self.bundle again if it's empty

        for key in raw.keys():
            field = camel_to_snake(key)
            self.bundle[field] = raw[key]

    def dispatch(self, request, *args, **kwargs):
        method = request.method.lower()
        handler = self.http_method_not_allowed

        if method in self.http_method_names:
            handler = getattr(self, method, self.http_method_not_allowed)

        try:
            self._check_permissions()  # only returns 200 or HTTP_EXCEPTIONS
            self._assemble_bundle_from_request_body()  # sets self.bundle
            return handler(request, *args, **kwargs)  # calls self.serialize()
        except HTTP_EXCEPTIONS as e:
            return self.render_to_response(dict(message=e.message), e.status)
        except ObjectDoesNotExist:
            return self.render_to_response(
                dict(message=HTTP404.message), HTTP404.status
            )
        except ValidationError as e:
            return self.render_to_response(dict(message=e.message), HTTP422.status)



from django.db import models


class DummyModel(models.Model):
    id = models.CharField(max_length=64, primary_key=True)
    email = models.CharField(max_length=64)
    phone = models.CharField(max_length=64)

    def api(self):
        return dict(id=self.id, email=self.email, phone=self.phone)

    def api_update_fields(self):
        return [
            "id",
            "email",
            "phone",
        ]
